# -*- coding: utf-8 -*-

# Copyright (c) 2012-2019, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.

from datetime import datetime

from pyramid.httpexceptions import HTTPInternalServerError, \
    HTTPNotFound, HTTPBadRequest, HTTPForbidden
from pyramid.view import view_config

from sqlalchemy import Enum, exc, func, distinct, Numeric, String, Text, Unicode, UnicodeText
from sqlalchemy.orm.exc import NoResultFound, MultipleResultsFound
from sqlalchemy.sql import and_, or_
from sqlalchemy.orm.util import class_mapper
from sqlalchemy.orm.properties import ColumnProperty

from geoalchemy2 import Geometry, func as ga_func
from geoalchemy2.shape import from_shape, to_shape

import geojson
from geojson.feature import FeatureCollection, Feature

from shapely.geometry import asShape
from shapely.ops import cascaded_union
from shapely.geos import TopologicalError

from papyrus.protocol import Protocol, create_filter
from papyrus.xsd import XSDGenerator

from c2cgeoportal_commons import models
from c2cgeoportal_commons.models.main import Layer, RestrictionArea, Role
from c2cgeoportal_geoportal.lib.caching import get_region, \
    set_common_headers, NO_CACHE, PUBLIC_CACHE, PRIVATE_CACHE
from c2cgeoportal_geoportal.lib.dbreflection import get_class, get_table, _AssociationProxy

import logging
log = logging.getLogger(__name__)

cache_region = get_region()


class Layers:

    def __init__(self, request):
        self.request = request
        self.settings = request.registry.settings.get("layers", {})
        self.layers_enum_config = self.settings.get("enum")

    @staticmethod
    def _get_geom_col_info(layer):
        """ Return information about the layer's geometry column, namely
        a ``(name, srid)`` tuple, where ``name`` is the name of the
        geometry column, and ``srid`` its srid.

        This function assumes that the names of geometry attributes
        in the mapped class are the same as those of geometry columns.
        """
        mapped_class = get_layer_class(layer)
        for p in class_mapper(mapped_class).iterate_properties:
            if not isinstance(p, ColumnProperty):
                continue  # pragma: no cover
            col = p.columns[0]
            if isinstance(col.type, Geometry):
                return col.name, col.type.srid
        raise HTTPInternalServerError(
            'Failed getting geometry column info for table "{0!s}".'.format(
                layer.geo_table
            )
        )  # pragma: no cover

    @staticmethod
    def _get_layer(layer_id):
        """ Return a ``Layer`` object for ``layer_id``. """
        layer_id = int(layer_id)
        try:
            query = models.DBSession.query(Layer, Layer.geo_table)
            query = query.filter(Layer.id == layer_id)
            layer, geo_table = query.one()
        except NoResultFound:
            raise HTTPNotFound("Layer {0:d} not found".format(layer_id))
        except MultipleResultsFound:  # pragma: no cover
            raise HTTPInternalServerError(
                "Too many layers found with id {0:d}".format(layer_id)
            )
        if not geo_table:  # pragma: no cover
            raise HTTPNotFound("Layer {0:d} has no geo table".format(layer_id))
        return layer

    def _get_layers_for_request(self):
        """ A generator function that yields ``Layer`` objects based
        on the layer ids found in the ``layer_id`` matchdict. """
        try:
            layer_ids = (
                int(layer_id) for layer_id in
                self.request.matchdict["layer_id"].split(",") if layer_id)
            for layer_id in layer_ids:
                yield self._get_layer(layer_id)
        except ValueError:
            raise HTTPBadRequest(
                "A Layer id in '{0!s}' is not an integer".format(
                    self.request.matchdict["layer_id"]
                )
            )  # pragma: no cover

    def _get_layer_for_request(self):
        """ Return a ``Layer`` object for the first layer id found
        in the ``layer_id`` matchdict. """
        return next(self._get_layers_for_request())

    def _get_protocol_for_layer(self, layer, **kwargs):
        """ Returns a papyrus ``Protocol`` for the ``Layer`` object. """
        cls = get_layer_class(layer)
        geom_attr = self._get_geom_col_info(layer)[0]
        return Protocol(models.DBSession, cls, geom_attr, **kwargs)

    def _get_protocol_for_request(self, **kwargs):
        """ Returns a papyrus ``Protocol`` for the first layer
        id found in the ``layer_id`` matchdict. """
        layer = self._get_layer_for_request()
        return self._get_protocol_for_layer(layer, **kwargs)

    def _proto_read(self, layer):
        """ Read features for the layer based on the self.request. """
        proto = self._get_protocol_for_layer(layer)
        if layer.public:
            return proto.read(self.request)
        user = self.request.user
        if user is None:
            raise HTTPForbidden()
        cls = proto.mapped_class
        geom_attr = proto.geom_attr
        ras = models.DBSession.query(RestrictionArea.area, RestrictionArea.area.ST_SRID())
        ras = ras.join(RestrictionArea.roles)
        ras = ras.join(RestrictionArea.layers)
        ras = ras.filter(Role.id.in_([role.id for role in user.roles]))
        ras = ras.filter(Layer.id == layer.id)
        collect_ra = []
        use_srid = -1
        for ra, srid in ras.all():
            if ra is None:
                return proto.read(self.request)
            else:
                use_srid = srid
                collect_ra.append(to_shape(ra))
        if len(collect_ra) == 0:  # pragma: no cover
            raise HTTPForbidden()

        filter1_ = create_filter(self.request, cls, geom_attr)
        ra = cascaded_union(collect_ra)
        filter2_ = ga_func.ST_Contains(
            from_shape(ra, use_srid),
            getattr(cls, geom_attr)
        )
        filter_ = filter2_ if filter1_ is None else and_(filter1_, filter2_)

        return proto.read(self.request, filter=filter_)

    @view_config(route_name="layers_read_many", renderer="geojson")
    def read_many(self):
        set_common_headers(self.request, "layers", NO_CACHE)

        features = []
        for layer in self._get_layers_for_request():
            for f in self._proto_read(layer).features:
                f.properties["__layer_id__"] = layer.id
                features.append(f)

        return FeatureCollection(features)

    @view_config(route_name="layers_read_one", renderer="geojson")
    def read_one(self):
        set_common_headers(self.request, "layers", NO_CACHE)

        layer = self._get_layer_for_request()
        protocol = self._get_protocol_for_layer(layer)
        feature_id = self.request.matchdict.get("feature_id")
        feature = protocol.read(self.request, id=feature_id)
        if not isinstance(feature, Feature):
            return feature
        if layer.public:
            return feature
        if self.request.user is None:
            raise HTTPForbidden()
        geom = feature.geometry
        if not geom or isinstance(geom, geojson.geometry.Default):  # pragma: no cover
            return feature
        shape = asShape(geom)
        srid = self._get_geom_col_info(layer)[1]
        spatial_elt = from_shape(shape, srid=srid)
        allowed = models.DBSession.query(func.count(RestrictionArea.id))
        allowed = allowed.join(RestrictionArea.roles)
        allowed = allowed.join(RestrictionArea.layers)
        allowed = allowed.filter(Role.id.in_([role.id for role in self.request.user.roles]))
        allowed = allowed.filter(Layer.id == layer.id)
        allowed = allowed.filter(or_(
            RestrictionArea.area.is_(None),
            RestrictionArea.area.ST_Contains(spatial_elt)
        ))
        if allowed.scalar() == 0:
            raise HTTPForbidden()

        return feature

    @view_config(route_name="layers_count", renderer="string")
    def count(self):
        set_common_headers(self.request, "layers", NO_CACHE)

        protocol = self._get_protocol_for_request()
        return protocol.count(self.request)

    @view_config(route_name="layers_create", renderer="geojson")
    def create(self):
        set_common_headers(self.request, "layers", NO_CACHE)

        if self.request.user is None:
            raise HTTPForbidden()

        self.request.response.cache_control.no_cache = True

        layer = self._get_layer_for_request()

        def check_geometry(_, feature, obj):
            del obj  # unused
            geom = feature.geometry
            if geom and not isinstance(geom, geojson.geometry.Default):
                shape = asShape(geom)
                srid = self._get_geom_col_info(layer)[1]
                spatial_elt = from_shape(shape, srid=srid)
                allowed = models.DBSession.query(func.count(RestrictionArea.id))
                allowed = allowed.join(RestrictionArea.roles)
                allowed = allowed.join(RestrictionArea.layers)
                allowed = allowed.filter(RestrictionArea.readwrite.is_(True))
                allowed = allowed.filter(Role.id.in_([r.id for r in self.request.user.roles]))
                allowed = allowed.filter(Layer.id == layer.id)
                allowed = allowed.filter(or_(
                    RestrictionArea.area.is_(None),
                    RestrictionArea.area.ST_Contains(spatial_elt)
                ))
                if allowed.scalar() == 0:
                    raise HTTPForbidden()

                # check if geometry is valid
                if self._get_validation_setting(layer):
                    self._validate_geometry(spatial_elt)

        protocol = self._get_protocol_for_layer(layer, before_create=check_geometry)
        try:
            features = protocol.create(self.request)
            for feature in features.features:
                self._log_last_update(layer, feature)
            return features
        except TopologicalError as e:
            self.request.response.status_int = 400
            return {"error_type": "validation_error",
                    "message": str(e)}
        except exc.IntegrityError as e:
            log.error(str(e))
            self.request.response.status_int = 400
            return {"error_type": "integrity_error",
                    "message": str(e.orig.diag.message_primary)}

    @view_config(route_name="layers_update", renderer="geojson")
    def update(self):
        set_common_headers(self.request, "layers", NO_CACHE)

        if self.request.user is None:
            raise HTTPForbidden()

        self.request.response.cache_control.no_cache = True

        feature_id = self.request.matchdict.get("feature_id")
        layer = self._get_layer_for_request()

        def check_geometry(_, feature, obj):
            # we need both the "original" and "new" geometry to be
            # within the restriction area
            geom_attr, srid = self._get_geom_col_info(layer)
            geom_attr = getattr(obj, geom_attr)
            geom = feature.geometry
            allowed = models.DBSession.query(func.count(RestrictionArea.id))
            allowed = allowed.join(RestrictionArea.roles)
            allowed = allowed.join(RestrictionArea.layers)
            allowed = allowed.filter(RestrictionArea.readwrite.is_(True))
            allowed = allowed.filter(Role.id.in_([role.id for role in self.request.user.roles]))
            allowed = allowed.filter(Layer.id == layer.id)
            allowed = allowed.filter(or_(
                RestrictionArea.area.is_(None),
                RestrictionArea.area.ST_Contains(geom_attr)
            ))
            spatial_elt = None
            if geom and not isinstance(geom, geojson.geometry.Default):
                shape = asShape(geom)
                spatial_elt = from_shape(shape, srid=srid)
                allowed = allowed.filter(or_(
                    RestrictionArea.area.is_(None),
                    RestrictionArea.area.ST_Contains(spatial_elt)
                ))
            if allowed.scalar() == 0:
                raise HTTPForbidden()

            # check is geometry is valid
            if self._get_validation_setting(layer):
                self._validate_geometry(spatial_elt)

        protocol = self._get_protocol_for_layer(layer, before_update=check_geometry)
        try:
            feature = protocol.update(self.request, feature_id)
            self._log_last_update(layer, feature)
            return feature
        except TopologicalError as e:
            self.request.response.status_int = 400
            return {"error_type": "validation_error",
                    "message": str(e)}
        except exc.IntegrityError as e:
            log.error(str(e))
            self.request.response.status_int = 400
            return {"error_type": "integrity_error",
                    "message": str(e.orig.diag.message_primary)}

    @staticmethod
    def _validate_geometry(geom):
        if geom is not None:
            simple = models.DBSession.query(func.ST_IsSimple(geom)).scalar()
            if not simple:
                raise TopologicalError("Not simple")
            valid = models.DBSession.query(func.ST_IsValid(geom)).scalar()
            if not valid:
                reason = models.DBSession.query(func.ST_IsValidReason(geom)).scalar()
                raise TopologicalError(reason)

    def _log_last_update(self, layer, feature):
        last_update_date = self.get_metadata(layer, "lastUpdateDateColumn")
        if last_update_date is not None:
            setattr(feature, last_update_date, datetime.now())

        last_update_user = self.get_metadata(layer, "lastUpdateUserColumn")
        if last_update_user is not None:
            setattr(feature, last_update_user, self.request.user.id)

    @staticmethod
    def get_metadata(layer, key, default=None):
        metadatas = layer.get_metadatas(key)
        if len(metadatas) == 1:
            metadata = metadatas[0]
            return metadata.value
        return default

    def _get_validation_setting(self, layer):
        # The validation UIMetadata is stored as a string, not a boolean
        should_validate = self.get_metadata(layer, "geometryValidation", None)
        if should_validate:
            return should_validate.lower() != "false"
        return self.settings.get("geometry_validation", False)

    @view_config(route_name="layers_delete")
    def delete(self):
        if self.request.user is None:
            raise HTTPForbidden()

        feature_id = self.request.matchdict.get("feature_id")
        layer = self._get_layer_for_request()

        def security_cb(_, obj):
            geom_attr = getattr(obj, self._get_geom_col_info(layer)[0])
            allowed = models.DBSession.query(func.count(RestrictionArea.id))
            allowed = allowed.join(RestrictionArea.roles)
            allowed = allowed.join(RestrictionArea.layers)
            allowed = allowed.filter(RestrictionArea.readwrite.is_(True))
            allowed = allowed.filter(Role.id.in_([role.id for role in self.request.user.roles]))
            allowed = allowed.filter(Layer.id == layer.id)
            allowed = allowed.filter(or_(
                RestrictionArea.area.is_(None),
                RestrictionArea.area.ST_Contains(geom_attr)
            ))
            if allowed.scalar() == 0:
                raise HTTPForbidden()

        protocol = self._get_protocol_for_layer(layer, before_delete=security_cb)
        response = protocol.delete(self.request, feature_id)
        set_common_headers(
            self.request, "layers", NO_CACHE,
            response=response,
        )
        return response

    @view_config(route_name="layers_metadata", renderer="xsd")
    def metadata(self):
        set_common_headers(self.request, "layers", PRIVATE_CACHE)

        layer = self._get_layer_for_request()
        if not layer.public and self.request.user is None:
            raise HTTPForbidden()

        return get_layer_class(layer, with_last_update_columns=True)

    @view_config(route_name="layers_enumerate_attribute_values", renderer="json")
    def enumerate_attribute_values(self):
        set_common_headers(self.request, "layers", PUBLIC_CACHE)

        if self.layers_enum_config is None:  # pragma: no cover
            raise HTTPInternalServerError("Missing configuration")
        layername = self.request.matchdict["layer_name"]
        fieldname = self.request.matchdict["field_name"]
        # TODO check if layer is public or not

        return self._enumerate_attribute_values(
            layername, fieldname
        )

    @cache_region.cache_on_arguments()
    def _enumerate_attribute_values(self, layername, fieldname):
        if layername not in self.layers_enum_config:  # pragma: no cover
            raise HTTPBadRequest("Unknown layer: {0!s}".format(layername))

        layerinfos = self.layers_enum_config[layername]
        if fieldname not in layerinfos["attributes"]:  # pragma: no cover
            raise HTTPBadRequest("Unknown attribute: {0!s}".format(fieldname))
        dbsession_name = layerinfos.get("dbsession", "dbsession")
        dbsession = models.DBSessions.get(dbsession_name)
        if dbsession is None:  # pragma: no cover
            raise HTTPInternalServerError(
                "No dbsession found for layer '{0!s}' ({1!s})".format(layername, dbsession_name)
            )
        values = self.query_enumerate_attribute_values(dbsession, layerinfos, fieldname)
        enum = {
            "items": [{"label": value[0], "value": value[0]} for value in values]
        }
        return enum

    @staticmethod
    def query_enumerate_attribute_values(dbsession, layerinfos, fieldname):
        attrinfos = layerinfos["attributes"][fieldname]
        table = attrinfos["table"]
        layertable = get_table(table, session=dbsession)
        column = attrinfos.get("column_name", fieldname)
        attribute = getattr(layertable.columns, column)
        # For instance if `separator` is a "," we consider that the column contains a
        # comma separate list of values e.g.: "value1,value2".
        if "separator" in attrinfos:
            separator = attrinfos["separator"]
            attribute = func.unnest(func.string_to_array(
                func.string_agg(attribute, separator), separator
            ))
        return dbsession.query(distinct(attribute)).order_by(attribute).all()


def get_layer_class(layer, with_last_update_columns=False):
    """
    Get the SQLAlchemy class to edit a GeoMapFish layer

    :param layer:
    :param with_last_update_columns: False to just have a class to access to the table and be able to
           modify the last_update_columns, True to have a correct class to build the UI
           (without the hidden column).
    :return: SQLAlchemy class
    """
    # Exclude the columns used to record the last features update
    exclude = [] if layer.exclude_properties is None else layer.exclude_properties.split(",")
    if with_last_update_columns:
        last_update_date = Layers.get_metadata(layer, "lastUpdateDateColumn")
        if last_update_date is not None:
            exclude.append(last_update_date)
        last_update_user = Layers.get_metadata(layer, "lastUpdateUserColumn")
        if last_update_user is not None:
            exclude.append(last_update_user)
    else:
        exclude = []

    m = Layers.get_metadata(layer, "editingAttributesOrder")
    attributes_order = m.split(',') if m else None
    m = Layers.get_metadata(layer, "readonlyAttributes")
    readonly_attributes = m.split(',') if m else None

    primary_key = Layers.get_metadata(layer, "geotablePrimaryKey")
    cls = get_class(
        str(layer.geo_table),
        exclude_properties=exclude,
        primary_key=primary_key,
        attributes_order=attributes_order,
        readonly_attributes=readonly_attributes
    )

    mapper = class_mapper(cls)
    column_properties = [p.key for p in mapper.iterate_properties if isinstance(p, ColumnProperty)]
    for attribute_name in attributes_order or []:
        if attribute_name not in column_properties:
            table = mapper.mapped_table
            log.warning(
                'Attribute "{}" does not exists in table "{}".\n'
                'Please correct metadata "editingAttributesOrder" in layer "{}" (id={}).\n'
                'Available attributes are: {}.'.format(
                    attribute_name,
                    '{}.{}'.format(table.schema, table.name),
                    layer.name,
                    layer.id,
                    ', '.join(column_properties)
                )
            )

    return cls


def get_layer_metadatas(layer):
    cls = get_layer_class(layer, with_last_update_columns=True)
    edit_columns = []

    for column_property in class_mapper(cls).iterate_properties:
        if isinstance(column_property, ColumnProperty):

            if len(column_property.columns) != 1:
                raise NotImplementedError  # pragma: no cover

            column = column_property.columns[0]

            # Exclude columns that are primary keys
            if not column.primary_key:
                properties = _convert_column_type(column.type)
                properties["name"] = column.key

                if column.nullable:
                    properties["nillable"] = True
                edit_columns.append(properties)
        else:
            for k, p in cls.__dict__.items():
                if not isinstance(p, _AssociationProxy):
                    continue

                relationship_property = class_mapper(cls) \
                    .get_property(p.target)
                target_cls = relationship_property.argument
                query = models.DBSession.query(getattr(target_cls, p.value_attr))
                properties = {}
                if column.nullable:
                    properties["nillable"] = True

                properties["name"] = k
                properties["restriction"] = "enumeration"
                properties["type"] = "xsd:string"
                properties["enumeration"] = []
                for value in query:
                    properties["enumeration"].append(value[0])

                edit_columns.append(properties)
    return edit_columns


def _convert_column_type(column_type):
    # SIMPLE_XSD_TYPES
    for cls, xsd_type in XSDGenerator.SIMPLE_XSD_TYPES.items():
        if isinstance(column_type, cls):
            return {"type": xsd_type}

    # Geometry type
    if isinstance(column_type, Geometry):
        geometry_type = column_type.geometry_type
        if geometry_type in XSDGenerator.SIMPLE_GEOMETRY_XSD_TYPES:
            xsd_type = XSDGenerator.SIMPLE_GEOMETRY_XSD_TYPES[geometry_type]
            return {"type": xsd_type, "srid": int(column_type.srid)}
        if geometry_type == "GEOMETRY":
            xsd_type = "gml:GeometryPropertyType"
            return {"type": xsd_type, "srid": int(column_type.srid)}

        raise NotImplementedError  # pragma: no cover

    # Enumeration type
    if isinstance(column_type, Enum):
        restriction = {}
        restriction["restriction"] = "enumeration"
        restriction["type"] = "xsd:string"
        restriction["enumeration"] = []
        for enum in column_type.enums:
            restriction["enumeration"].append(enum)
        return restriction

    # String type
    if isinstance(column_type, (String, Text, Unicode, UnicodeText)):
        if column_type.length is None:
            return {"type": "xsd:string"}
        else:
            return {"type": "xsd:string", "maxLength": int(column_type.length)}

    # Numeric Type
    if isinstance(column_type, Numeric):
        xsd_type = {"type": "xsd:decimal"}
        if column_type.scale is None and column_type.precision is None:
            return xsd_type

        else:
            if column_type.scale is not None:
                xsd_type["fractionDigits"] = int(column_type.scale)
            if column_type.precision is not None:
                xsd_type["totalDigits"] = int(column_type.precision)
            return xsd_type

    raise NotImplementedError  # pragma: no cover



# -*- coding: utf-8 -*-

# Copyright (c) 2011-2019, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


import json
import logging
import re
import sys
import urllib.parse
import xml.dom.minidom  # noqa # pylint: disable=unused-import
from collections import Counter
from math import sqrt
from random import Random
from typing import Dict, Set, Tuple  # noqa # pylint: disable=unused-import

import requests
import zope.event.classhandler
from c2cwsgiutils.auth import auth_view
from defusedxml import lxml
from owslib.wms import WebMapService
from pyramid.httpexceptions import HTTPBadGateway, HTTPBadRequest, HTTPForbidden, HTTPFound
from pyramid.i18n import TranslationStringFactory
from pyramid.path import AssetResolver
from pyramid.renderers import render_to_response
from pyramid.response import Response
from pyramid.security import forget, remember
from pyramid.view import view_config
from sqlalchemy.orm import subqueryload
from sqlalchemy.orm.exc import NoResultFound

from c2cgeoportal_commons import models
from c2cgeoportal_commons.lib.email_ import send_email_config
from c2cgeoportal_commons.models import main, static
from c2cgeoportal_geoportal import is_valid_referer
from c2cgeoportal_geoportal.lib import add_url_params, get_setting, get_typed, get_types_map, get_url2
from c2cgeoportal_geoportal.lib.cacheversion import get_cache_version
from c2cgeoportal_geoportal.lib.caching import (
    NO_CACHE, PRIVATE_CACHE, PUBLIC_CACHE, get_region, set_common_headers)
from c2cgeoportal_geoportal.lib.functionality import get_functionality, get_mapserver_substitution_params
from c2cgeoportal_geoportal.lib.layers import get_protected_layers_query
from c2cgeoportal_geoportal.lib.wmstparsing import TimeInformation, parse_extent
from c2cgeoportal_geoportal.views.layers import get_layer_metadatas

_ = TranslationStringFactory("c2cgeoportal")
LOG = logging.getLogger(__name__)
CACHE_REGION = get_region()


class DimensionInformation:

    URL_PART_RE = re.compile(r"[a-zA-Z0-9_\-\+~\.]*$")

    def __init__(self):
        self._dimensions = {}
        self._dimension_filters = {}

    def merge(self, layer, layer_node, mixed):
        errors = set()

        dimensions = {}
        dimensions_filters = {}
        for dimension in layer.dimensions:
            if not isinstance(layer, main.LayerWMS) and dimension.value is not None and \
                    not self.URL_PART_RE.match(dimension.value):
                errors.add(u"The layer '{}' has an unsupported dimension value '{}' ('{}').".format(
                    layer.name, dimension.value, dimension.name
                ))
            elif dimension.name in dimensions:  # pragma: nocover
                errors.add("The layer '{}' has a duplicated dimension name '{}'.".format(
                    layer.name, dimension.name
                ))
            else:
                if dimension.field:
                    dimensions_filters[dimension.name] = {
                        'field': dimension.field,
                        'value': dimension.value
                    }
                else:
                    dimensions[dimension.name] = dimension.value

        if dimensions_filters:
            layer_node["dimensionsFilters"] = dimensions_filters
        if mixed:
            layer_node["dimensions"] = dimensions
        else:
            for name, value in list(dimensions.items()):
                if name not in self._dimensions or self._dimensions[name] is None:
                    self._dimensions[name] = value
                elif self._dimensions[name] != value and value is not None:
                    errors.add(
                        "The layer '{}' has a wrong dimension value '{}' for '{}', "
                        "expected '{}' or empty.".format(
                            layer.name, value, name, self._dimensions[name]
                        )
                    )
        return errors

    def get_dimensions(self):
        return self._dimensions


class Entry:

    WFS_NS = "http://www.opengis.net/wfs"
    server_wms_capabilities = {}  # type: Dict[int, Tuple[WebMapService, Set[str]]]
    server_wfs_feature_type = {}  # type: Dict[int, xml.dom.minidom.Document]

    def __init__(self, request):
        self.request = request
        self.settings = request.registry.settings
        self.http_options = self.request.registry.settings.get("http_options", {})
        self.debug = "debug" in request.params
        self.lang = request.locale_name
        self.metadata_type = get_types_map(
            self.settings.get("admin_interface", {}).get("available_metadata", [])
        )
        self._layerswms_cache = None
        self._layerswmts_cache = None
        self._layergroup_cache = None
        self._themes_cache = None

        from c2cgeoportal_commons.models.main import InvalidateCacheEvent

        @zope.event.classhandler.handler(InvalidateCacheEvent)
        def handle(event: InvalidateCacheEvent):  # pylint: disable=unused-variable
            del event
            Entry.server_wms_capabilities = {}
            Entry.server_wfs_feature_type = {}

    @view_config(route_name="testi18n", renderer="testi18n.html")
    def testi18n(self):  # pragma: no cover
        _ = self.request.translate
        return {"title": _("title i18n")}

    def _get_capabilities_cache_role_key(self, ogc_server):
        return self._get_role_ids() if (
            ogc_server.auth != main.OGCSERVER_AUTH_NOAUTH
            and ogc_server.type != main.OGCSERVER_TYPE_MAPSERVER
        ) else None

    def _get_metadata(self, item, metadata, errors):
        metadatas = item.get_metadatas(metadata)
        return \
            None if len(metadatas) == 0 \
            else get_typed(
                metadata, metadatas[0].value,
                self.metadata_type, self.request, errors,
                layer_name=item.name
            )

    def _get_metadatas(self, item, errors):
        metadatas = {}
        for metadata in item.metadatas:
            value = get_typed(
                metadata.name, metadata.value,
                self.metadata_type, self.request, errors
            )
            if value is not None:
                metadatas[metadata.name] = value

        return metadatas

    def _wms_getcap(self, ogc_server):
        if ogc_server.id in self.server_wms_capabilities:
            return self.server_wms_capabilities[ogc_server.id]

        url, content, errors = self._wms_getcap_cached(
            ogc_server, self._get_capabilities_cache_role_key(ogc_server)
        )

        if len(errors) != 0:
            return None, errors

        wms = None
        try:
            wms = WebMapService(None, xml=content)
        except Exception:  # pragma: no cover
            error = "WARNING! an error occurred while trying to read the mapfile and recover the themes." \
                "\nURL: {}\n{}".format(url, content)
            errors.add(error)
            LOG.error(error, exc_info=True)

        self.server_wms_capabilities[ogc_server.id] = (wms, errors)

        return wms, errors

    @CACHE_REGION.cache_on_arguments()
    def get_http_cached(self, url, headers):
        return requests.get(url, headers=headers, **self.http_options)

    @CACHE_REGION.cache_on_arguments()
    def _wms_getcap_cached(self, ogc_server, _):
        """ _ is just for cache on the role id """

        errors = set()
        url = get_url2(
            "The OGC server '{}'".format(ogc_server.name),
            ogc_server.url, self.request, errors
        )
        if len(errors):  # pragma: no cover
            return url, None, errors

        # Add functionality params
        sparams = get_mapserver_substitution_params(self.request)
        url = add_url_params(url, sparams)

        errors = set()

        url = add_url_params(url, {
            "SERVICE": "WMS",
            "VERSION": "1.1.1",
            "REQUEST": "GetCapabilities",
            "ROLE_ID": "0",
            "USER_ID": "0",
        })

        LOG.info("Get WMS GetCapabilities for url: %s", url)

        # Forward request to target (without Host Header)
        headers = dict(self.request.headers)

        # Add headers for Geoserver
        if ogc_server.auth == main.OGCSERVER_AUTH_GEOSERVER and self.request.user is not None:
            headers["sec-username"] = self.request.user.username
            headers["sec-roles"] = "" if self.request.user is None else \
                ";".join([role.name for role in self.request.user.roles])

        if urllib.parse.urlsplit(url).hostname != "localhost" and "Host" in headers:  # pragma: no cover
            headers.pop("Host")

        try:
            response = self.get_http_cached(url, headers)
        except Exception:  # pragma: no cover
            error = "Unable to GetCapabilities from URL {}".format(url)
            errors.add(error)
            LOG.error(error, exc_info=True)
            return url, None, errors

        if not response.ok:  # pragma: no cover
            error = "GetCapabilities from URL {} return the error: {:d} {}".format(
                url, response.status_code, response.reason
            )
            errors.add(error)
            LOG.error(error)
            return url, None, errors

        # With wms 1.3 it returns text/xml also in case of error :-(
        if response.headers["Content-Type"].split(";")[0].strip() not in \
                ["application/vnd.ogc.wms_xml", "text/xml"]:
            error = "GetCapabilities from URL {} returns a wrong Content-Type: {}\n{}".format(
                url, response.headers["Content-Type"], response.text
            )
            errors.add(error)
            LOG.error(error)
            return url, None, errors

        return url, response.content, errors

    def _create_layer_query(self, interface):
        """
        Create an SQLAlchemy query for Layer and for the role
        identified to by ``role_id``.
        """

        query = models.DBSession.query(main.Layer.name).filter(
            main.Layer.item_type.notin_(["layerv1"])
        )

        query = query.filter(main.Layer.public.is_(True))

        if interface is not None:
            query = query.join(main.Layer.interfaces)
            query = query.filter(main.Interface.name == interface)

        if self.request.user is not None:
            query2 = get_protected_layers_query(self.request.user, None, what=main.LayerWMS.name)
            if interface is not None:
                query2 = query2.join(main.Layer.interfaces)
                query2 = query2.filter(main.Interface.name == interface)
            query = query.union(query2)
            query3 = get_protected_layers_query(self.request.user, None, what=main.LayerWMTS.name)
            if interface is not None:
                query3 = query3.join(main.Layer.interfaces)
                query3 = query3.filter(main.Interface.name == interface)
            query = query.union(query3)

        return query

    def _get_layer_metadata_urls(self, layer):
        metadata_urls = []
        if len(layer.metadataUrls) > 0:
            metadata_urls = layer.metadataUrls
        for child_layer in layer.layers:
            metadata_urls.extend(self._get_layer_metadata_urls(child_layer))
        return metadata_urls

    def _get_layer_resolution_hint_raw(self, layer):
        resolution_hint_min = None
        resolution_hint_max = None
        if layer.scaleHint:
            # scaleHint is based upon a pixel diagonal length whereas we use
            # resolutions based upon a pixel edge length. There is a sqrt(2)
            # ratio between edge and diagonal of a square.
            resolution_hint_min = float(layer.scaleHint["min"]) / sqrt(2)
            resolution_hint_max = float(layer.scaleHint["max"]) / sqrt(2) \
                if layer.scaleHint["max"] != "0" else 999999999
        for child_layer in layer.layers:
            resolution = self._get_layer_resolution_hint_raw(child_layer)
            resolution_hint_min = resolution[0] if resolution_hint_min is None else (
                resolution_hint_min if resolution[0] is None else
                min(resolution_hint_min, resolution[0])
            )
            resolution_hint_max = resolution[1] if resolution_hint_max is None else (
                resolution_hint_max if resolution[1] is None else
                max(resolution_hint_max, resolution[1])
            )

        return (resolution_hint_min, resolution_hint_max)

    def _get_layer_resolution_hint(self, layer):
        resolution_hint_min, resolution_hint_max = self._get_layer_resolution_hint_raw(layer)
        return (
            0.0 if resolution_hint_min is None else resolution_hint_min,
            999999999 if resolution_hint_max is None else resolution_hint_max,
        )

    def _get_child_layers_info_1(self, layer):
        """ Return information about sub layers of a layer.

            Arguments:

            * ``layer`` The layer object in the WMS capabilities.
        """
        child_layers_info = []
        for child_layer in layer.layers:
            resolution = self._get_layer_resolution_hint(child_layer)
            child_layer_info = {
                "name": child_layer.name,
                "minResolutionHint": float("{:0.2f}".format(resolution[0])),
                "maxResolutionHint": float("{:0.2f}".format(resolution[1])),
            }
            if hasattr(child_layer, "queryable"):
                child_layer_info["queryable"] = child_layer.queryable
            child_layers_info.append(child_layer_info)
        return child_layers_info

    def _get_child_layers_info(self, layer):
        """ Return information about sub layers of a layer.

            Arguments:

            * ``layer`` The layer object in the WMS capabilities.
        """
        resolution = self._get_layer_resolution_hint(layer)
        layer_info = {
            "name": layer.name,
            "minResolutionHint": float("{:0.2f}".format(resolution[0])),
            "maxResolutionHint": float("{:0.2f}".format(resolution[1])),
        }
        layer_info["queryable"] = layer.queryable == 1 \
            if hasattr(layer, "queryable") else True
        return layer_info

    def _layer(self, layer, time=None, dim=None, mixed=True):
        errors = set()
        layer_info = {
            "id": layer.id,
            "name": layer.name,
            "metadata": self._get_metadatas(layer, errors),
        }
        if re.search("[/?#]", layer.name):  # pragma: no cover
            errors.add("The layer has an unsupported name '{}'.".format(layer.name))
        if isinstance(layer, main.LayerWMS) and re.search("[/?#]", layer.layer):  # pragma: no cover
            errors.add("The layer has an unsupported layers '{}'.".format(layer.layer))
        if layer.geo_table:
            errors |= self._fill_editable(layer_info, layer)
        if mixed:
            assert time is None
            time = TimeInformation()
        assert time is not None

        errors |= dim.merge(layer, layer_info, mixed)

        if isinstance(layer, main.LayerWMS):
            wms, wms_layers, wms_errors = self._wms_layers(layer.ogc_server)
            errors |= wms_errors
            if layer.layer is None or layer.layer == "":
                errors.add("The layer '{}' do not have any layers".format(layer.name))
                return None, errors
            layer_info["type"] = "WMS"
            layer_info["layers"] = layer.layer
            self._fill_wms(layer_info, layer, errors, mixed=mixed)
            errors |= self._merge_time(time, layer_info, layer, wms, wms_layers)

        elif isinstance(layer, main.LayerWMTS):
            layer_info["type"] = "WMTS"
            self._fill_wmts(layer_info, layer, errors)

        return layer_info, errors

    @staticmethod
    def _merge_time(time, l, layer, wms, wms_layers):
        errors = set()
        wmslayer = layer.layer
        try:
            if wmslayer in wms_layers:
                wms_layer_obj = wms[wmslayer]

                if wms_layer_obj.timepositions:
                    extent = parse_extent(
                        wms_layer_obj.timepositions,
                        wms_layer_obj.defaulttimeposition
                    )
                    time.merge(l, extent, layer.time_mode, layer.time_widget)

                for child_layer in wms_layer_obj.layers:
                    if child_layer.timepositions:
                        extent = parse_extent(
                            child_layer.timepositions,
                            child_layer.defaulttimeposition
                        )
                        # The time mode comes from the layer group
                        time.merge(l, extent, layer.time_mode, layer.time_widget)

        except ValueError:  # pragma no cover
            errors.add(
                "Error while handling time for layer '{}': {}".format(layer.name, sys.exc_info()[1])
            )

        return errors

    def _fill_editable(self, l, layer):
        errors = set()
        try:
            if self.request.user:
                count = models.DBSession.query(main.RestrictionArea) \
                    .join(main.RestrictionArea.roles) \
                    .filter(main.Role.id.in_([r.id for r in self.request.user.roles])) \
                    .filter(main.RestrictionArea.layers.any(main.Layer.id == layer.id)) \
                    .filter(main.RestrictionArea.readwrite.is_(True)) \
                    .count()
                if count > 0:
                    l["edit_columns"] = get_layer_metadatas(layer)
                    l["editable"] = True
        except Exception as exception:
            LOG.exception(str(exception))
            errors.add(str(exception))
        return errors

    def _fill_wms(self, l, layer, errors, mixed):
        wms, wms_layers, wms_errors = self._wms_layers(layer.ogc_server)
        errors |= wms_errors

        l["imageType"] = layer.ogc_server.image_type
        if layer.style:  # pragma: no cover
            l["style"] = layer.style

        # now look at what is in the WMS capabilities doc
        l["childLayers"] = []
        for layer_name in layer.layer.split(","):
            if layer_name in wms_layers:
                wms_layer_obj = wms[layer_name]
                if len(wms_layer_obj.layers) == 0:
                    l["childLayers"].append(self._get_child_layers_info(wms_layer_obj))
                else:
                    for child_layer in wms_layer_obj.layers:
                        l["childLayers"].append(self._get_child_layers_info(child_layer))
            else:
                errors.add(
                    "The layer '{}' ({}) is not defined in WMS capabilities from '{}'".format(
                        layer_name, layer.name, layer.ogc_server.name
                    )
                )

        if "minResolutionHint" not in l:
            resolution_min = self._get_metadata(layer, "minResolution", errors)

            if resolution_min is not None:
                l["minResolutionHint"] = resolution_min
            else:
                min_resolutions_hint = [
                    l_["minResolutionHint"]
                    for l_ in l["childLayers"]
                    if "minResolutionHint" in l_
                ]
                if len(min_resolutions_hint) > 0:
                    l["minResolutionHint"] = min(min_resolutions_hint)
        if "maxResolutionHint" not in l:
            resolution_max = self._get_metadata(layer, "maxResolution", errors)

            if resolution_max is not None:
                l["maxResolutionHint"] = resolution_max
            else:
                max_resolutions_hint = [
                    l_["maxResolutionHint"]
                    for l_ in l["childLayers"]
                    if "maxResolutionHint" in l_
                ]
                if len(max_resolutions_hint) > 0:
                    l["maxResolutionHint"] = max(max_resolutions_hint)

        if mixed:
            l["ogcServer"] = layer.ogc_server.name

    @staticmethod
    def _fill_legend_rule_query_string(l, layer, url):
        if layer.legend_rule and url:
            l["icon"] = add_url_params(url, {
                "SERVICE": "WMS",
                "VERSION": "1.1.1",
                "REQUEST": "GetLegendGraphic",
                "LAYER": layer.name,
                "FORMAT": "image/png",
                "TRANSPARENT": "TRUE",
                "RULE": layer.legend_rule,
            })

    def _fill_wmts(self, l, layer, errors):
        l["url"] = get_url2(
            "The WMTS layer '{}'".format(layer.name),
            layer.url, self.request, errors=errors
        )

        if layer.style:
            l["style"] = layer.style
        if layer.matrix_set:
            l["matrixSet"] = layer.matrix_set

        l["layer"] = layer.layer
        l["imageType"] = layer.image_type

    @staticmethod
    def _layer_included(tree_item):
        return isinstance(tree_item, main.Layer)

    def _get_ogc_servers(self, group, depth):
        """ Recurse on all children to get unique identifier for each child. """
        ogc_servers = set()

        # escape loop
        if depth > 30:
            LOG.error("Error: too many recursions with group '%s'", group.name)
            return ogc_servers

        # recurse on children
        if isinstance(group, main.LayerGroup) and len(group.children) > 0:
            for tree_item in group.children:
                ogc_servers.update(self._get_ogc_servers(tree_item, depth + 1))

        if isinstance(group, main.LayerWMS):
            ogc_servers.add(group.ogc_server.name)

        if isinstance(group, main.LayerWMTS):
            ogc_servers.add(False)

        return ogc_servers

    def _group(
        self, path, group, layers, depth=1, min_levels=1, mixed=True, time=None,
        dim=None, wms_layers=None, layers_name=None, **kwargs
    ):
        if wms_layers is None:
            wms_layers = []
        if layers_name is None:
            layers_name = []
        children = []
        errors = set()

        if re.search("[/?#]", group.name):  # pragma: no cover
            errors.add("The group has an unsupported name '{}'.".format(group.name))

        # escape loop
        if depth > 30:
            errors.add(
                "Too many recursions with group '{}'".format(group.name)
            )
            return None, errors

        ogc_servers = None
        org_depth = depth
        if depth == 1:
            ogc_servers = list(self._get_ogc_servers(group, 1))
            # check if mixed content
            mixed = len(ogc_servers) != 1 or ogc_servers[0] is False
            if not mixed:
                time = TimeInformation()
            dim = DimensionInformation()

        for tree_item in group.children:
            if isinstance(tree_item, main.LayerGroup):
                gp, gp_errors = self._group(
                    u"{}/{}".format(path, tree_item.name), tree_item, layers,
                    depth=depth + 1, min_levels=min_levels,
                    mixed=mixed, time=time, dim=dim, wms_layers=wms_layers, layers_name=layers_name, **kwargs
                )
                errors |= gp_errors
                if gp is not None:
                    children.append(gp)
            elif isinstance(tree_item, main.LayerV1):
                pass
            elif self._layer_included(tree_item):
                if tree_item.name in layers:
                    layers_name.append(tree_item.name)
                    if isinstance(tree_item, main.LayerWMS):
                        wms_layers.extend(tree_item.layer.split(","))

                    l, l_errors = self._layer(tree_item, mixed=mixed, time=time, dim=dim, **kwargs)
                    errors |= l_errors
                    if l is not None:
                        if depth < min_levels:
                            errors.add("The Layer '{}' is under indented ({:d}/{:d}).".format(
                                path + "/" + tree_item.name, depth, min_levels
                            ))
                        else:
                            children.append(l)

        if len(children) > 0:
            g = {
                "id": group.id,
                "name": group.name,
                "children": children,
                "metadata": self._get_metadatas(group, errors),
                "mixed": False,
            }
            if not mixed:
                for name, nb in list(Counter(layers_name).items()):
                    if nb > 1:
                        errors.add(
                            "The GeoMapFish layer name '{}', cannot be two times "
                            "in the same block (first level group).".format(name)
                        )

            g["mixed"] = mixed
            if org_depth == 1:
                if not mixed:
                    g["ogcServer"] = ogc_servers[0]
                    if time.has_time() and time.layer is None:
                        g["time"] = time.to_dict()

                    g["dimensions"] = dim.get_dimensions()

            return g, errors
        else:
            return None, errors

    def _layers(self, interface):
        query = self._create_layer_query(interface=interface)
        return [name for (name,) in query.all()]

    def _wms_layers(self, ogc_server):
        # retrieve layers metadata via GetCapabilities
        wms, wms_errors = self._wms_getcap(ogc_server)
        if len(wms_errors) > 0:
            return None, [], wms_errors

        return wms, list(wms.contents), wms_errors

    def _load_tree_items(self):
        self._layerswms_cache = models.DBSession.query(main.LayerWMS).options(
            subqueryload(main.LayerWMS.dimensions), subqueryload(main.LayerWMS.metadatas)
        ).all()
        self._layerswmts_cache = models.DBSession.query(main.LayerWMTS).options(
            subqueryload(main.LayerWMTS.dimensions), subqueryload(main.LayerWMTS.metadatas)
        ).all()
        self._layergroup_cache = models.DBSession.query(main.LayerGroup).options(
            subqueryload(main.LayerGroup.metadatas), subqueryload(main.LayerGroup.children_relation)
        ).all()
        self._themes_cache = models.DBSession.query(main.Theme).options(
            subqueryload(main.Theme.functionalities),
            subqueryload(main.Theme.metadatas),
            subqueryload(main.Theme.children_relation)
        ).all()

    def _themes(self, interface="desktop", filter_themes=True, min_levels=1):
        """
        This function returns theme information for the role identified
        by ``role_id``.
        """
        self._load_tree_items()
        errors = set()
        layers = self._layers(interface)

        themes = models.DBSession.query(main.Theme)
        themes = themes.filter(main.Theme.public.is_(True))
        if self.request.user is not None:
            auth_themes = models.DBSession.query(main.Theme)
            auth_themes = auth_themes.filter(main.Theme.public.is_(False))
            auth_themes = auth_themes.join(main.Theme.restricted_roles)
            auth_themes = auth_themes.filter(main.Role.id.in_([r.id for r in self.request.user.roles]))

            themes = themes.union(auth_themes)

        themes = themes.order_by(main.Theme.ordering.asc())

        if filter_themes and interface is not None:
            themes = themes.join(main.Theme.interfaces)
            themes = themes.filter(main.Interface.name == interface)

        export_themes = []
        for theme in themes.all():
            if re.search("[/?#]", theme.name):
                errors.add("The theme has an unsupported name '{}'.".format(theme.name))
                continue

            children, children_errors = self._get_children(theme, layers, min_levels)
            errors |= children_errors

            # test if the theme is visible for the current user
            if len(children) > 0:
                icon = get_url2(
                    "The Theme '{}'".format(theme.name),
                    theme.icon, self.request, errors,
                ) if theme.icon is not None and len(theme.icon) > 0 else self.request.static_url(
                    "{}_geoportal:static/images/blank.png".format(self.request.registry.settings["package"])
                )

                t = {
                    "id": theme.id,
                    "name": theme.name,
                    "icon": icon,
                    "children": children,
                    "functionalities": self._get_functionalities(theme),
                    "metadata": self._get_metadatas(theme, errors),
                }
                export_themes.append(t)

        return export_themes, errors

    @staticmethod
    def _get_functionalities(theme):
        result = {}
        for functionality in theme.functionalities:
            if functionality.name in result:
                result[functionality.name].append(functionality.value)
            else:
                result[functionality.name] = [functionality.value]
        return result

    @view_config(route_name="invalidate", renderer="json")
    def invalidate_cache(self):  # pragma: no cover
        auth_view(self.request)
        main.cache_invalidate_cb()
        return {
            "success": True
        }

    def _get_children(self, theme, layers, min_levels):
        children = []
        errors = set()
        for item in theme.children:
            if isinstance(item, main.LayerGroup):
                gp, gp_errors = self._group(
                    "{}/{}".format(theme.name, item.name), item, layers, min_levels=min_levels
                )
                errors |= gp_errors
                if gp is not None:
                    children.append(gp)
            elif self._layer_included(item):
                if min_levels > 0:
                    errors.add("The Layer '{}' cannot be directly in the theme '{}' (0/{:d}).".format(
                        item.name, theme.name, min_levels
                    ))
                elif item.name in layers:
                    l, l_errors = self._layer(item, dim=DimensionInformation())
                    errors |= l_errors
                    if l is not None:
                        children.append(l)
        return children, errors

    def _functionality(self):
        functionality = {}
        for func_ in get_setting(
                self.settings,
                ("functionalities", "available_in_templates"), []
        ):
            functionality[func_] = get_functionality(
                func_, self.request
            )
        return functionality

    @CACHE_REGION.cache_on_arguments()
    def _get_layers_enum(self):
        layers_enum = {}
        if "enum" in self.settings.get("layers", {}):
            for layer_name, layer in \
                    list(self.settings["layers"]["enum"].items()):
                layer_enum = {}
                layers_enum[layer_name] = layer_enum
                for attribute in list(layer["attributes"].keys()):
                    layer_enum[attribute] = self.request.route_url(
                        "layers_enumerate_attribute_values",
                        layer_name=layer_name,
                        field_name=attribute,
                        path=""
                    )
        return layers_enum

    def _get_role_ids(self):
        return None if self.request.user is None else \
            {role.id for role in self.request.user.roles}

    def get_ngeo_index_vars(self):
        set_common_headers(self.request, "index", NO_CACHE, content_type="text/html")
        return {}

    def get_oldapijs_values(self):
        ogc_server = models.DBSession.query(main.OGCServer).filter(
            main.OGCServer.name == self.settings["api"]["ogc_server"]
        ).one()
        wms, wms_errors = self._wms_getcap(ogc_server)
        if len(wms_errors) > 0:  # pragma: no cover
            raise HTTPBadGateway("\n".join(wms_errors))
        queryable_layers = [
            name for name in list(wms.contents)
            if wms[name].queryable == 1]
        cache_version = get_cache_version()
        return {
            "lang": self.lang,
            "debug": self.debug,
            "queryable_layers": json.dumps(queryable_layers),
            "url_params": {"cache_version": cache_version} if cache_version else {},
            "tiles_url": json.dumps(self.settings.get("tiles_url")),
            "request": self.request,
        }

    @view_config(route_name="apijs")
    def apijs(self):
        version = self.request.params.get('version', '1')
        if version == '2':
            resolver = AssetResolver('c2cgeoportal_geoportal')
            asset = resolver.resolve('{}_geoportal:static-ngeo/build/api.js'.format(
                self.request.registry.settings['package']
            ))
            self.request.response.body = asset.stream().read()
            response = self.request.response
        else:
            response = render_to_response(
                'api/api.js', self.get_oldapijs_values()
            )
        set_common_headers(
            self.request, "api", PUBLIC_CACHE,
            response=response, content_type="application/javascript",
        )
        return response

    @view_config(route_name="oldxapijs", renderer="api/xapi.js")
    def oldxapijs(self):
        ogc_server = models.DBSession.query(main.OGCServer).filter(
            main.OGCServer.name == self.settings["api"]["ogc_server"]
        ).one()
        wms, wms_errors = self._wms_getcap(ogc_server)
        if len(wms_errors) > 0:  # pragma: no cover
            raise HTTPBadGateway("\n".join(wms_errors))
        queryable_layers = [
            name for name in list(wms.contents)
            if wms[name].queryable == 1
        ]
        cache_version = get_cache_version()

        set_common_headers(
            self.request, "api", NO_CACHE,
            content_type="application/javascript",
        )

        return {
            "lang": self.lang,
            "debug": self.debug,
            "queryable_layers": json.dumps(queryable_layers),
            "url_params": {"cache_version": cache_version} if cache_version else {},
            "tiles_url": json.dumps(self.settings.get("tiles_url")),
        }

    @view_config(route_name="oldapihelp", renderer="api/apihelp.html")
    def oldapihelp(self):
        set_common_headers(self.request, "index", NO_CACHE)

        return {
            "lang": self.lang,
            "debug": self.debug,
        }

    @view_config(route_name="oldxapihelp", renderer="api/xapihelp.html")
    def oldxapihelp(self):
        set_common_headers(self.request, "index", NO_CACHE)

        return {
            "lang": self.lang,
            "debug": self.debug,
        }

    def favicon(self):
        set_common_headers(
            self.request, 'api', NO_CACHE,
            content_type='image/vnd.microsoft.icon',
        )
        return {}

    def apijsmap(self):
        set_common_headers(
            self.request, "api", NO_CACHE,
            content_type="application/octet-stream",
        )
        return {}

    def apicss(self):
        set_common_headers(
            self.request, "api", PUBLIC_CACHE,
            content_type="text/css",
        )
        return {}

    def apihelp(self):
        set_common_headers(self.request, "index", NO_CACHE)
        return {}

    def _wms_get_features_type(self, ogc_server_id, wfs_url):
        errors = set()

        if ogc_server_id in self.server_wfs_feature_type:
            return self.server_wfs_feature_type[ogc_server_id], errors

        params = {
            "SERVICE": "WFS",
            "VERSION": "1.0.0",
            "REQUEST": "DescribeFeatureType",
        }
        wfs_url = add_url_params(wfs_url, params)

        LOG.info("WFS DescribeFeatureType for base url: %s", wfs_url)

        # forward request to target (without Host Header)
        headers = dict(self.request.headers)
        if urllib.parse.urlsplit(wfs_url).hostname != "localhost" and "Host" in headers:
            headers.pop("Host")  # pragma nocover

        try:
            response = self.get_http_cached(wfs_url, headers)
        except Exception:  # pragma: no cover
            errors.add("Unable to get DescribeFeatureType from URL {}".format(wfs_url))
            return None, errors

        if not response.ok:  # pragma: no cover
            errors.add("DescribeFeatureType from URL {} return the error: {:d} {}".format(
                wfs_url, response.status_code, response.reason
            ))
            return None, errors

        try:
            self.server_wfs_feature_type[ogc_server_id] = lxml.XML(response.text.encode('utf-8'))
            return self.server_wfs_feature_type[ogc_server_id], errors
        except Exception as e:  # pragma: no cover
            errors.add("Error '{}' on reading DescribeFeatureType from URL {}:\n{}".format(
                str(e), wfs_url, response.text
            ))
            return None, errors

    @view_config(route_name="themes", renderer="json")
    def themes(self):
        interface = self.request.params.get("interface", "desktop")
        sets = self.request.params.get("set", "all")
        min_levels = int(self.request.params.get("min_levels", 1))
        group = self.request.params.get("group")
        background_layers_group = self.request.params.get("background")

        export_themes = sets in ("all", "themes")
        export_group = group is not None and sets in ("all", "group")
        export_background = background_layers_group is not None and sets in ("all", "background")

        set_common_headers(self.request, "themes", PRIVATE_CACHE)

        result = {}
        all_errors = set()
        result["ogcServers"] = {}
        for ogc_server in models.DBSession.query(main.OGCServer).all():
            # required to do every time to validate the url.
            if ogc_server.auth != main.OGCSERVER_AUTH_NOAUTH:
                url = self.request.route_url("mapserverproxy", _query={"ogcserver": ogc_server.name})
                url_wfs = url
            else:
                url = get_url2(
                    "The OGC server '{}'".format(ogc_server.name),
                    ogc_server.url, self.request, errors=all_errors
                )
                url_wfs = get_url2(
                    "The OGC server (WFS) '{}'".format(ogc_server.name),
                    ogc_server.url_wfs, self.request, errors=all_errors
                ) if ogc_server.url_wfs is not None else url
            feature_type, errors = self._wms_get_features_type(ogc_server.id, url_wfs)
            all_errors |= errors
            attributes = None

            if feature_type is not None:
                types = {}
                elements = {}
                for child in feature_type.getchildren():
                    if child.tag == '{http://www.w3.org/2001/XMLSchema}element':
                        elements[child.attrib['name']] = child.attrib['type'].split(':')[1]

                    if child.tag == '{http://www.w3.org/2001/XMLSchema}complexType':
                        s = child.find('.//{http://www.w3.org/2001/XMLSchema}sequence')
                        attrib = {}
                        for c in s.getchildren():
                            namespace = None
                            type_ = c.attrib['type']
                            if len(type_.split(':')) == 2:
                                namespace, type_ = type_.split(':')
                            namespace = c.nsmap[namespace]
                            attrib[c.attrib['name']] = {
                                'namespace': namespace,
                                'type': type_,
                            }
                            if 'alias' in c.attrib:
                                attrib[c.attrib['name']] = c.attrib['alias']
                        types[child.attrib['name']] = attrib
                attributes = {}
                for name, type_ in elements.items():
                    if type_ in types:
                        attributes[name] = types[type_]
                    else:
                        LOG.warning(
                            "The provided type '%s' does not exist, available types are %s.",
                            type_, ', '.join(types.keys())
                        )
            result["ogcServers"][ogc_server.name] = {
                "url": url,
                "urlWfs": url_wfs,
                "type": ogc_server.type,
                "credential": ogc_server.auth != main.OGCSERVER_AUTH_NOAUTH,
                "imageType": ogc_server.image_type,
                "wfsSupport": ogc_server.wfs_support,
                "isSingleTile": ogc_server.is_single_tile,
                "namespace": feature_type.attrib["targetNamespace"]
                if feature_type is not None else None,
                "attributes": attributes,
            }
        if export_themes:
            themes, errors = self._themes(interface, True, min_levels)

            result["themes"] = themes
            all_errors |= errors

        if export_group:
            group, errors = self._get_group(group, interface)
            if group is not None:
                result["group"] = group
            all_errors |= errors

        if export_background:
            group, errors = self._get_group(background_layers_group, interface)
            result["background_layers"] = group["children"] if group is not None else []
            all_errors |= errors

        result["errors"] = list(all_errors)
        if len(all_errors) > 0:
            LOG.info("Theme errors:\n%s", "\n".join(all_errors))
        return result

    def _get_group(self, group, interface):
        layers = self._layers(interface)
        try:
            lg = models.DBSession.query(main.LayerGroup).filter(main.LayerGroup.name == group).one()
            return self._group(lg.name, lg, layers)
        except NoResultFound:  # pragma: no cover
            return None, set(["Unable to find the Group named: {}, Available Groups: {}".format(
                group, ", ".join([i[0] for i in models.DBSession.query(main.LayerGroup.name).all()])
            )])

    def _referer_log(self):
        if not hasattr(self.request, "is_valid_referer"):
            self.request.is_valid_referer = is_valid_referer(self.request)
        if not self.request.is_valid_referer:
            LOG.error(
                "Invalid referer for %s: %s", self.request.path_qs, repr(self.request.referer)
            )

    @view_config(context=HTTPForbidden, renderer="login.html")
    def loginform403(self):
        self._referer_log()

        if self.request.authenticated_userid:
            return HTTPForbidden()  # pragma: no cover

        set_common_headers(self.request, "login", NO_CACHE)

        return {
            "lang": self.lang,
            "came_from": self.request.path,
        }

    @view_config(route_name="loginform", renderer="login.html")
    def loginform(self):
        self._referer_log()

        set_common_headers(self.request, "login", PUBLIC_CACHE, vary=True)

        return {
            "lang": self.lang,
            "came_from": self.request.params.get("came_from") or "/",
        }

    @view_config(route_name="login")
    def login(self):
        self._referer_log()

        login = self.request.POST.get("login")
        password = self.request.POST.get("password")
        if login is None or password is None:  # pragma nocover
            LOG.info("'login' and 'password' should be available in request params.")
            raise HTTPBadRequest("See server logs for details")
        username = self.request.registry.validate_user(self.request, login, password)
        if username is not None:
            user = models.DBSession.query(static.User).filter(static.User.username == username).one_or_none()
            if user:
                user.update_last_login()
            headers = remember(self.request, username)
            LOG.info("User '%s' logged in.", username)
            came_from = self.request.params.get("came_from")
            if came_from:
                return HTTPFound(location=came_from, headers=headers)
            else:
                headers.append(("Content-Type", "text/json"))
                return set_common_headers(
                    self.request, "login", NO_CACHE,
                    response=Response(json.dumps(self._user(
                        self.request.get_user(username)
                    )), headers=headers),
                )
        else:
            raise HTTPForbidden("See server logs for details")

    @view_config(route_name="logout")
    def logout(self):
        headers = forget(self.request)

        if not self.request.user:
            LOG.info("Logout on non login user.")
            raise HTTPBadRequest("See server logs for details")

        LOG.info("User '%s' (%s) logging out.", self.request.user.username, self.request.user.id)

        headers.append(("Content-Type", "text/json"))
        return set_common_headers(
            self.request, "login", NO_CACHE,
            response=Response("true", headers=headers),
        )

    def _user(self, user=None):
        user = self.request.user if user is None else user
        result = {
            "success": True,  # for Extjs
            "username": user.username,
            "email": user.email,
            "is_password_changed": user.is_password_changed,
            "roles": [{"name": r.name, "id": r.id} for r in user.roles],
        } if user else {}

        result["functionalities"] = self._functionality()

        return result

    @view_config(route_name="loginuser", renderer="json")
    def loginuser(self):
        self._referer_log()

        set_common_headers(self.request, "login", NO_CACHE)

        return self._user()

    @view_config(route_name="loginchange", renderer="json")
    def loginchange(self):
        set_common_headers(self.request, "login", NO_CACHE)

        old_password = self.request.POST.get("oldPassword")
        new_password = self.request.POST.get("newPassword")
        new_password_confirm = self.request.POST.get("confirmNewPassword")
        if new_password is None or new_password_confirm is None or old_password is None:
            LOG.info(
                "'oldPassword', 'newPassword' and 'confirmNewPassword' should be "
                "available in request params."
            )
            raise HTTPBadRequest("See server logs for details")

        # check if logged in
        if not self.request.user:  # pragma nocover
            LOG.info("Change password on non login user.")
            raise HTTPBadRequest("See server logs for details")

        user = self.request.registry.validate_user(
            self.request, self.request.user.username, old_password
        )
        if user is None:
            LOG.info("The old password is wrong for user '%s'.", user)
            raise HTTPBadRequest("See server logs for details")

        if new_password != new_password_confirm:
            LOG.info(
                "The new password and the new password "
                "confirmation do not match for user '%s'.", user
            )
            raise HTTPBadRequest("See server logs for details")

        u = self.request.user
        u.password = new_password
        u.is_password_changed = True
        models.DBSession.flush()
        LOG.info("Password changed for user '%s'", self.request.user.username)

        return {
            "success": "true"
        }

    @staticmethod
    def generate_password():
        allchars = "123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
        rand = Random()

        password = ""
        for _ in range(8):
            password += rand.choice(allchars)

        return password

    def _loginresetpassword(self):
        username = self.request.POST["login"]
        try:
            user = models.DBSession.query(static.User).filter(static.User.username == username).one()
        except NoResultFound:  # pragma: no cover
            return None, None, None, "The login '{}' does not exist.".format(username)

        if user.email is None or user.email == "":  # pragma: no cover
            return \
                None, None, None, \
                "The user '{}' has no registered email address.".format(user.username),

        password = self.generate_password()
        user.set_temp_password(password)

        return user, username, password, None

    @view_config(route_name="loginresetpassword", renderer="json")
    def loginresetpassword(self):  # pragma: no cover
        set_common_headers(
            self.request, "login", NO_CACHE
        )

        user, username, password, error = self._loginresetpassword()
        if error is not None:
            LOG.info(error)
            raise HTTPBadRequest("See server logs for details")

        send_email_config(
            self.request.registry.settings, "reset_password", user.email,
            user=username, password=password
        )

        return {
            "success": True
        }
